@@ -1,14 +1,14 @@
-"""Objects that support some list-like read operations on an unbounded stream.
+"""
+Objects that support some list-like read operations on an unbounded stream.
 Essentially, trying to give you the impression that you have read access to infinite list,
 with some (parametrizable) limitations.
 
 """
 # TODO: Build up extensive relations expression and handling, but InfiniteSeq only uses BEFORE (past).
 #  Consider simplifying.
-# TODO:
 
 from collections import deque
-from typing import Iterable, Tuple, Union, Callable
+from typing import Iterable, Tuple, Union, Callable, Any, NewType
 from functools import wraps, partial, partialmethod
 from enum import Enum
 from operator import le, lt, ge, gt, itemgetter
@@ -32,7 +32,9 @@ class Relations(Enum):
     See Allen's interval algebra for (some of the) interval relations
     (https://en.wikipedia.org/wiki/Allen%27s_interval_algebra).
     """
-    # simple relations, that can be used between (X: point, Y: interval), (X: interval, Y: interval) pairs
+
+    # simple relations, that can be used between
+    # (X: point, Y: interval) or (X: interval, Y: interval) pairs
     BEFORE = 'Some of X happens BEFORE Y'
     DURING = 'All of X happens within Y'
     AFTER = 'Some of X happens AFTER Y'
@@ -53,14 +55,16 @@ def validate_interval(interval):
         assert bt <= tt
         return bt, tt
     except Exception as e:
-        raise ValueError(f"Not a valid interval: {interval}")
+        raise ValueError(f'Not a valid interval: {interval}')
 
 
 # TODO: Validate intervals (assert x[0] <= x[1] and )?
-def simple_interval_relationship(x: Tuple[Number, Number],
-                                 y: Tuple[Number, Number],
-                                 above_bt: Callable = ge,
-                                 below_tt: Callable = lt):
+def simple_interval_relationship(
+    x: Tuple[Number, Number],
+    y: Tuple[Number, Number],
+    above_bt: Callable = ge,
+    below_tt: Callable = lt,
+):
     """Get the simple relationship between intervals x and y.
 
     :param x: An point (a number) or an interval (a 2-tuple of numbers).
@@ -72,9 +76,10 @@ def simple_interval_relationship(x: Tuple[Number, Number],
         Relations.AFTER if some of x is after y,
         Relations.DURING if x is entirely with y
 
-    The target ``y`` interval is expressed only by it's bounds, but we don't know if these are
-     inclusive or not. The ``below_bt`` and ``above_tt`` allow us to express that by expressing how below the
-     lowest (bt) bound and what higher than highest (tt) bound are defined.
+    The target ``y`` interval is expressed only by it's bounds, but we don't know if
+     these are inclusive or not. The ``below_bt`` and ``above_tt`` allow us to express
+     that by expressing how below the lowest (bt) bound and what higher than highest
+     (tt) bound are defined.
 
     The function is meant to be curried (partial), for example:
 
@@ -84,8 +89,10 @@ def simple_interval_relationship(x: Tuple[Number, Number],
     >>> including_bounds = partial(simple_interval_relationship, above_bt=ge, below_tt=le)
     >>> excluding_bounds = partial(simple_interval_relationship, above_bt=gt, below_tt=lt)
 
-    Take ``(4, 8)`` as the target interval, and want to query the relationship of other points and intervals with it.
-    No matter what the function is, they will always agree on any intervals that don't share any bounds.
+    Take ``(4, 8)`` as the target interval, and want to query the relationship of other
+    points and intervals with it.
+    No matter what the function is, they will always agree on any intervals that don't
+    share any bounds.
 
     >>> for relation_func in (default, including_bounds, excluding_bounds):
     ...     print (
@@ -135,9 +142,9 @@ def simple_interval_relationship(x: Tuple[Number, Number],
     else:
         x_bt, x_tt = x, x  # consider a point to be the (x, x) interval
     y_bt, y_tt = validate_interval(y)
-    if not above_bt(x_bt, y_bt):
+    if not above_bt(x_bt, y_bt):  # meaning x_bt > y_bt (or >=, depending on above_bt)
         return Relations.BEFORE
-    elif below_tt(x_tt, y_tt):
+    elif below_tt(x_tt, y_tt):  # meaning x_bt < y_tt (or <=, depending on below_tt)
         return Relations.DURING
     else:
         return Relations.AFTER
@@ -147,6 +154,7 @@ def simple_interval_relationship(x: Tuple[Number, Number],
 class ExceptionRaiserCallbackMixin:
     """Make the instance callable and have the effect of raising the instance.
     Meant to add to an exception class so that instances of this class can be used as callbacks that raise the error"""
+
     dflt_args = ()
 
     def __init__(self, *args, **kwargs):
@@ -165,19 +173,22 @@ class ExceptionRaiserCallbackMixin:
 class NotDuringError(ExceptionRaiserCallbackMixin, IndexError):
     """IndexError that indicates that there was an attempt to index some data that is not contained in the buffer
     (i.e. is that a part of the request is NO LONGER, or NOT YET covered by the buffer)"""
-    dflt_args = "Some of the data requested was in the past or in the future"
+
+    dflt_args = 'Some of the data requested was in the past or in the future'
 
 
 class OverlapsPastError(NotDuringError):
     """IndexError that indicates that there was an attempt to index some data that is in the PAST
     (i.e. is NO LONGER completely covered by the buffer)"""
-    dlft_args = "Some of the data requested is in the past"
+
+    dlft_args = 'Some of the data requested is in the past'
 
 
 class OverlapsFutureError(NotDuringError):
     """IndexError that indicates that there was an attempt to index some data that is in the FUTURE
     (i.e. is NOT YET completely covered by the buffer)"""
-    dlft_args = "Some of the data requested is in the future"
+
+    dlft_args = 'Some of the data requested is in the future'
 
 
 class RelationNotHandledError(TypeError):
@@ -193,7 +204,7 @@ overlaps_future_error = OverlapsFutureError()
 
 
 def _not_implemented(self, method_name, *args, **kwargs):
-    raise NotImplementedError("")
+    raise NotImplementedError('')
 
 
 # ram heavier, cpu lighter extend
@@ -236,7 +247,9 @@ def shift_slice(slice_obj, shift: Number):
 
 
 def absolute_item(item, max_idx):
-    """Returns an item with absolute references: i.e. with negative indices idx resolved to max_idx + idx
+    """Returns an item with absolute references: i.e. with negative indices idx
+    resolved to max_idx + idx
+
     >>> absolute_item(-1, 10)
     9
     >>> absolute_item(slice(-4, -2, 2), 10)
@@ -262,18 +275,19 @@ def absolute_item(item, max_idx):
         return item
 
 
-# TODO: Make operations thread safe
 class IndexedBuffer:
     """A list-like object that gives a limited-past read view of an unbounded stream
 
-    For example, say we had the stream of increasing integers 0, 1, 2, ... that is being fed to indexedBuffer
+    For example, say we had the stream of increasing integers 0, 1, 2, ...
+    that is being fed to indexedBuffer
 
-    What IndexedBuffer(maxlen=4) offers is access to the buffer's contents, but using the indices that
+    What IndexedBuffer(maxlen=4) offers is access to the buffer's contents,
+    but using the indices that
     the stream (if it were one big list in memory) would use instead of the buffer's index.
         0 1 2 3 [4 5 6 7] 8 9
 
-    IndexedBuffer uses collections.deque, exposing the append, extend, and clear methods, updating the index reference
-    in a thread-safe manner.
+    IndexedBuffer uses collections.deque, exposing the append, extend,
+    and clear methods, updating the index reference in a thread-safe manner.
 
     >>> s = IndexedBuffer(buffer_len=4)
     >>> s.extend(range(4))  # adding 4 elements in bulk (filling the buffer completely)
@@ -333,10 +347,14 @@ class IndexedBuffer:
     OverlapsFutureError: You asked for slice(4, 9, None), but the buffer only contains the index range: 2:6
     """
 
-    def __init__(self, buffer_len, prefill=(),
-                 if_overlaps_past=overlaps_past_error,
-                 if_overlaps_future=overlaps_future_error,
-                 slice_get_postproc: Callable = list):
+    def __init__(
+        self,
+        buffer_len,
+        prefill=(),
+        if_overlaps_past=overlaps_past_error,
+        if_overlaps_future=overlaps_future_error,
+        slice_get_postproc: Callable = list,
+    ):
         self._deque = deque(prefill, buffer_len)
         self.buffer_len = self._deque.maxlen
         self.max_idx = 0  # should correspond to the number of items added
@@ -346,7 +364,12 @@ class IndexedBuffer:
         self._lock = Lock()
 
     def __repr__(self):
-        return f"{type(self).__name__}(buffer_len={self.buffer_len}, min_idx={self.min_idx}, max_idx={self.max_idx}, ...)"
+        return (
+            f'{type(self).__name__}('
+            f'buffer_len={self.buffer_len}, '
+            f'min_idx={self.min_idx}, '
+            f'max_idx={self.max_idx}, ...)'
+        )
 
     def __iter__(self):
         yield from self._deque
@@ -367,11 +390,17 @@ class IndexedBuffer:
         elif isinstance(idx, Iterable):
             return tuple(x - self.min_idx for x in idx)
         else:
-            raise TypeError(f"{type(idx)} are not handled. You requested the outer_to_buffer_idx of {idx}")
+            raise TypeError(
+                f'{type(idx)} are not handled. You requested the outer_to_buffer_idx of {idx}'
+            )
 
     def __getitem__(self, item):
-        item = absolute_item(item, self.max_idx)  # Note: Overhead for convenience of negative indices use (worth it?)
-        relationship = simple_interval_relationship(item, (self.min_idx, self.max_idx + 1))
+        item = absolute_item(
+            item, self.max_idx
+        )  # Note: Overhead for convenience of negative indices use (worth it?)
+        relationship = simple_interval_relationship(
+            item, (self.min_idx, self.max_idx + 1)
+        )
         if relationship == Relations.DURING:
             item = self.outer_to_buffer_idx(item)
             if isinstance(item, slice):
@@ -389,15 +418,19 @@ class IndexedBuffer:
         elif relationship == Relations.BEFORE:
             raise self._overlaps_past_error(item)
         else:
-            raise RelationNotHandledError(f"The relation {relationship} is not handled.")
+            raise RelationNotHandledError(
+                f'The relation {relationship} is not handled.'
+            )
 
     def _overlaps_past_error(self, item):
         return OverlapsPastError(
-            f"You asked for {item}, but the buffer only contains the index range: {self.min_idx}:{self.max_idx}")
+            f'You asked for {item}, but the buffer only contains the index range: {self.min_idx}:{self.max_idx}'
+        )
 
     def _overlaps_future_error(self, item):
         return OverlapsFutureError(
-            f"You asked for {item}, but the buffer only contains the index range: {self.min_idx}:{self.max_idx}")
+            f'You asked for {item}, but the buffer only contains the index range: {self.min_idx}:{self.max_idx}'
+        )
 
     def append(self, x) -> None:
         with self._lock:
@@ -430,16 +463,36 @@ from dataclasses import dataclass
 from typing import Iterator
 
 
+# TODO: Add some mechanism to deal with ITERABLE instead of just iterator. As it is we have some unwanted behavior with
+#   iterables
 @dataclass
 class InfiniteSeq:
     """A list-like (read) view of an unbounded sequence/stream.
 
+    It is the combination of `IndexedBuffer` and an iterator that will be used to
+    source the buffer according to the slices that are requested.
+
+    If a slice is requested whose data is "in the future", the iterator will be
+    consumed until the buffer can satisfy that request.
+    If the requested slice has any part of it that is "in the past", that is,
+    has already been iterated through and is not in the buffer anymore, a
+    `OverlapsPastError` will be raised.
+
+    Therefore, `InfiniteSeq` is meant for ordered slice queries of size no more than
+    the buffer size.
+    If these conditions are satisfied, an `InfiniteSeq` will behave (with `i:j`
+    queries) as if it were one long list in memory.
+
+    Can be used with a live stream of data as long as the buffer size is big enough
+    to handle the data production and query rates.
+
     For example, take an iterator that cycles from 0 to 99 forever:
 
     >>> from itertools import cycle
     >>> iterator = cycle(range(100))
 
-    Let's make an InfiniteSeq instance for this stream, accomodating for a view of up to 11 items.
+    Let's make an `InfiniteSeq` instance for this stream, accomodating for a view of
+    up to 11 items.
 
     >>> s = InfiniteSeq(iterator, buffer_len=11)
 
@@ -470,11 +523,23 @@ class InfiniteSeq:
     [95, 96, 97, 98, 99, 0, 1, 2, 3, 4]
 
     You can also use slices with step and with negative integers (referencing the head of the buffer)
+
     >>> s[120:130:2]
     [20, 22, 24, 26, 28]
+    >>> s[120:130]
+    [20, 21, 22, 23, 24, 25, 26, 27, 28, 29]
     >>> s[-8:-2]
     [22, 23, 24, 25, 26, 27]
 
+    but you cannot slice farther back than the buffer
+
+    >>> try:
+    ...     s[-20:-2]
+    ... except OverlapsPastError as e:
+    ...     msg_text = str(e)
+    >>> print(msg_text)
+    You asked for slice(110, 128, None), but the buffer only contains the index range: 119:130
+
     Sometimes the source provides data in chunks. Sometimes these chunks are not even of fixed size.
     In those situations, you can use ``itertools.chain`` to "flatten" the iterator as in the following example:
 
@@ -533,6 +598,7 @@ class InfiniteSeq:
     [40, 41]
 
     """
+
     iterator: Iterator
     buffer_len: int
 
@@ -543,9 +609,128 @@ class InfiniteSeq:
         if isinstance(item, slice):
             n_ticks_in_the_future = item.stop - self.indexed_buffer.max_idx
             if n_ticks_in_the_future > 0:
-                # TODO: If indexed_buffer had a "fast-forward" (perhaps "peek") it could wasted less buffer writes
+                # TODO: If indexed_buffer had a "fast-forward" (perhaps "peek")
+                #  we could waste less buffer writes
                 self.indexed_buffer.extend(islice(self.iterator, n_ticks_in_the_future))
                 # consume(self.iterator, n_ticks_in_the_future)
             return self.indexed_buffer[item]
         elif isinstance(item, int):
             return self[slice(item, item + 1)][0]
+
+
+def new_type(name, typ, doc=None):
+    t = NewType(name, type)
+    if doc is not None:
+        t.__doc__ = doc
+    return t
+
+
+BufferInput = new_type(
+    'BufferInput', Any, 'input_of/what_we_insert_in a buffer (before transformation)'
+)
+
+BufferItem = new_type(
+    'BufferItem', Any, 'An item of a buffer (after input is transformed)'
+)
+
+InputDataTrans = new_type(
+    'InputDataTrans',
+    Callable[[BufferInput], BufferItem],
+    'A function that transforms a BufferInput in to a BufferItem',
+)
+Query = new_type('Query', Any, 'A query (i.e. key, selection specification, etc.)')
+QueryTrans = new_type(
+    'QueryTrans',
+    Callable[[Query], Query],
+    'A function transforming a query into another, ready to be applied form',
+)
+
+FiltFunc = Callable[[BufferItem], bool]
+
+
+def asis(obj):
+    return obj
+
+
+# TODO: Finish up and document
+class BufferedGetter:
+    """
+    `BufferedGetter` is intended to be a more general (but not optimized) class that
+    offers a query-interface to a buffer, intended to be used when the buffer is
+    being filled by a (possibly live) stream of data items.
+
+    By contrast...
+    The `IndexedBuffer` is a particular case where the queries are slices and the index
+    that is sliced on is an enumeration one.
+    The `InfiniteSeq` is a class combining `IndexedBuffer` with a data source it can
+    pull data from (according to the demands of the query).
+
+    >>> from creek.infinite_sequence import BufferedGetter
+    >>>
+    >>>
+    >>> b = BufferedGetter(20)
+    >>> b.extend([
+    ...     (1, 3, 'completely before'),
+    ...     (2, 4, 'still completely before (upper bounds are strict)'),
+    ...     (3, 6, 'partially before, but overlaps bottom'),
+    ...     (4, 5, 'totally', 'inside'),  # <- note this tuple has 4 elements
+    ...     (5, 8),  # <- note this tuple has only the minimum (2) elements,
+    ...     (7, 10, 'partially after, but overlaps top'),
+    ...     (8, 11, 'completely after (strict upper bound)'),
+    ...     (100, 101, 'completely after (obviously)')
+    ... ])
+    >>> b[lambda x: 3 < x[0] < 8]  # doctest: +NORMALIZE_WHITESPACE
+    [(4, 5, 'totally', 'inside'),
+     (5, 8),
+     (7, 10, 'partially after, but overlaps top')]
+    """
+
+    def __init__(
+        self,
+        buffer_len,
+        prefill=(),
+        input_data_trans=asis,
+        query_trans=asis,
+        # if_overlaps_past=overlaps_past_error,
+        # if_overlaps_future=overlaps_future_error,
+        slice_get_postproc: Callable = list,
+    ):
+        self._deque = deque(prefill, buffer_len)
+        self.buffer_len = self._deque.maxlen
+        self.input_data_trans = input_data_trans
+        self.query_trans = query_trans
+        # self.max_idx = 0  # should correspond to the number of items added
+        # self.if_overlaps_past = if_overlaps_past
+        # self.if_overlaps_future = if_overlaps_future
+        self.slice_get_postproc = slice_get_postproc
+        self._lock = Lock()
+
+    def ingress(self, x: BufferInput) -> BufferItem:
+        return x
+
+    def append(self, x: BufferInput) -> None:
+        with self._lock:
+            x = self.input_data_trans(x)
+            self._deque.append(x)
+
+    def extend(self, iterable):
+        for x in iterable:
+            self.append(x)
+
+    def clear(self):
+        with self._lock:
+            self._deque.clear()
+
+    def __getitem__(self, q: Query):
+        return self._getitem(q)
+
+    def _getitem(self, q: Query):
+        # note: just a composition
+        return self.slice_get_postproc(self.filter(self.query_trans(q)))
+
+    def filter(self, filt: Callable):
+        # assert callable(filt), f'filt must be callable, was: {filt}'
+        return filter(filt, self._deque)
+
+    def __iter__(self):  # to dodge the iteration falling back to __getitem__(i)
+        yield from self._deque
